@@ -52,6 +52,9 @@ typedef struct {
     ngx_int_t                   charset;
     ngx_int_t                   source_charset;
     ngx_flag_t                  override_charset;
+
+    ngx_hash_t                  types;
+    ngx_array_t                *types_keys;
 } ngx_http_charset_loc_conf_t;
 
 
@@ -110,6 +113,17 @@ static char *ngx_http_charset_merge_loc_conf(ngx_conf_t *cf,
 static ngx_int_t ngx_http_charset_postconfiguration(ngx_conf_t *cf);
 
 
+ngx_str_t  ngx_http_charset_default_types[] = {
+    ngx_string("text/html"),
+    ngx_string("text/xml"),
+    ngx_string("text/plain"),
+    ngx_string("text/vnd.wap.wml"),
+    ngx_string("application/x-javascript"),
+    ngx_string("application/rss+xml"),
+    ngx_null_string
+};
+
+
 static ngx_command_t  ngx_http_charset_filter_commands[] = {
 
     { ngx_string("charset"),
@@ -136,6 +150,13 @@ static ngx_command_t  ngx_http_charset_filter_commands[] = {
       offsetof(ngx_http_charset_loc_conf_t, override_charset),
       NULL },
 
+    { ngx_string("charset_types"),
+      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_1MORE,
+      ngx_http_types_slot,
+      NGX_HTTP_LOC_CONF_OFFSET,
+      offsetof(ngx_http_charset_loc_conf_t, types_keys),
+      &ngx_http_charset_default_types[0] },
+
     { ngx_string("charset_map"),
       NGX_HTTP_MAIN_CONF|NGX_CONF_BLOCK|NGX_CONF_TAKE2,
       ngx_http_charset_map_block,
@@ -185,7 +206,6 @@ static ngx_http_output_body_filter_pt    ngx_http_next_body_filter;
 static ngx_int_t
 ngx_http_charset_header_filter(ngx_http_request_t *r)
 {
-    u_char                        *ct;
     ngx_int_t                      charset, source_charset;
     ngx_str_t                     *mc, *from, *to, s;
     ngx_uint_t                     n;
@@ -204,7 +224,8 @@ ngx_http_charset_header_filter(ngx_http_request_t *r)
 
     if (r == r->main) {
 
-        if (r->headers_out.content_encoding
+        if (!r->ignore_content_encoding
+            && r->headers_out.content_encoding
             && r->headers_out.content_encoding->value.len)
         {
             return ngx_http_next_header_filter(r);
@@ -243,13 +264,7 @@ ngx_http_charset_header_filter(ngx_http_request_t *r)
                 }
 
             } else {
-                ct = r->headers_out.content_type.data;
-
-                if (ngx_strncasecmp(ct, (u_char *) "text/", 5) != 0
-                    && ngx_strncasecmp(ct,
-                                      (u_char *) "application/x-javascript", 24)
-                       != 0)
-                {
+                if (ngx_http_test_content_type(r, &mlcf->types) == NULL) {
                     return ngx_http_next_header_filter(r);
                 }
             }
@@ -527,12 +542,6 @@ ngx_http_charset_body_filter(ngx_http_request_t *r, ngx_chain_t *in)
                 break;
             }
 
-#if (NGX_HAVE_WRITE_ZEROCOPY)
-            if (b->zerocopy_busy) {
-                break;
-            }
-#endif
-
             ctx->busy = cl->next;
 
             if (b->tag != (ngx_buf_tag_t) &ngx_http_charset_filter_module) {
@@ -642,7 +651,7 @@ ngx_http_charset_recode_from_utf8(ngx_pool_t *pool, ngx_buf_t *buf,
                 size = buf->last - src;
 
                 saved = src;
-                n = ngx_utf_decode(&saved, size);
+                n = ngx_utf8_decode(&saved, size);
 
                 if (n == 0xfffffffe) {
                     /* incomplete UTF-8 symbol */
@@ -710,7 +719,7 @@ ngx_http_charset_recode_from_utf8(ngx_pool_t *pool, ngx_buf_t *buf,
     }
 
     saved = ctx->saved;
-    n = ngx_utf_decode(&saved, i);
+    n = ngx_utf8_decode(&saved, i);
 
     c = '\0';
 
@@ -818,7 +827,7 @@ recode:
 
         len = buf->last - src;
 
-        n = ngx_utf_decode(&src, len);
+        n = ngx_utf8_decode(&src, len);
 
         if (n < 0x10000) {
 
@@ -1270,7 +1279,7 @@ ngx_http_charset_map(ngx_conf_t *cf, ngx_command_t *dummy, void *conf)
 
         p = &table->src2dst[src * NGX_UTF_LEN] + 1;
 
-        n = ngx_utf_decode(&p, i);
+        n = ngx_utf8_decode(&p, i);
 
         if (n > 0xffff) {
             ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
@@ -1413,19 +1422,21 @@ ngx_http_charset_create_main_conf(ngx_conf_t *cf)
     }
 
     if (ngx_array_init(&mcf->charsets, cf->pool, 2, sizeof(ngx_http_charset_t))
-        == NGX_ERROR)
+        != NGX_OK)
     {
         return NGX_CONF_ERROR;
     }
 
     if (ngx_array_init(&mcf->tables, cf->pool, 1,
-                       sizeof(ngx_http_charset_tables_t)) == NGX_ERROR)
+                       sizeof(ngx_http_charset_tables_t))
+        != NGX_OK)
     {
         return NGX_CONF_ERROR;
     }
 
     if (ngx_array_init(&mcf->recodes, cf->pool, 2,
-                       sizeof(ngx_http_charset_recode_t)) == NGX_ERROR)
+                       sizeof(ngx_http_charset_recode_t))
+        != NGX_OK)
     {
         return NGX_CONF_ERROR;
     }
@@ -1444,6 +1455,13 @@ ngx_http_charset_create_loc_conf(ngx_conf_t *cf)
         return NGX_CONF_ERROR;
     }
 
+    /*
+     * set by ngx_pcalloc():
+     *
+     *     lcf->types = { NULL };
+     *     lcf->types_keys = NULL;
+     */
+
     lcf->charset = NGX_CONF_UNSET;
     lcf->source_charset = NGX_CONF_UNSET;
     lcf->override_charset = NGX_CONF_UNSET;
@@ -1462,6 +1480,14 @@ ngx_http_charset_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
     ngx_http_charset_recode_t     *recode;
     ngx_http_charset_main_conf_t  *mcf;
 
+    if (ngx_http_merge_types(cf, conf->types_keys, &conf->types,
+                             prev->types_keys, &prev->types,
+                             ngx_http_charset_default_types)
+        != NGX_OK)
+    {
+        return NGX_CONF_ERROR;
+    }
+
     ngx_conf_merge_value(conf->override_charset, prev->override_charset, 0);
     ngx_conf_merge_value(conf->charset, prev->charset, NGX_HTTP_NO_CHARSET);
 
